import Copyright.Copyright;
import java.util.Vector;
import java.util.Enumeration;

import java.awt.*;
import java.awt.event.*;

/**
 * This is the class that does all the interaction and display work. Get input from the user and allow them
 * to interact with the data (add points move them around and designate a point as a target point).
 * To get a smooth display without flickering during screen update this class uses double buffering (data is
 * rendered to an offscreen image and then copied to screen.
 * @author Ziv Yaniv (zivy@isis.georgetown.edu)
 */
public class Point2DManipulationPanel extends Panel {
    /*
     * Constants used for displaying the point sets, transformed and original,
     * and the cooridnate axis manipulator.
     */
    private final int HALF_MARK_SIZE = 3;
    private final int MARK_SIZE = 2*HALF_MARK_SIZE;
    private final int HALF_AXIS_CIRCLE_SIZE = 4;
    private final int AXIS_CIRCLE_SIZE = 2*HALF_AXIS_CIRCLE_SIZE;
    private final int AXIS_SIZE = 30;
    private Image offScreenImage; // Image used for double buffering.

    /*
     * Point data and coordinate axis data
     */
                 //a vector of PointPair's
    private Vector points;     
                          //origin, COG of the manipulated data
    private Point2D centerOfGravity;
                         //tip of the y axis (circle's center) 
    private Point2D yAxis;  
                        //tip of the x axis (circle's center) 
    private Point2D xAxis;  

    /*
     * Reference to object used for receiving input points.
     * Once the transformed set of points are manipulated this reference
     * is detached from the applet, disallowing further entries.
     */ 
    private MouseInputClicks mouseClicks;

    public Point2DManipulationPanel() {
              //add MouseListener instance which receives the input points
    mouseClicks = new MouseInputClicks();
    addMouseListener(mouseClicks);
         
             //add MouseMotion instance which manipulates the input points
    MouseInputDrags mia = new MouseInputDrags();
    addMouseMotionListener(mia);
    addMouseListener(mia);

    addComponentListener(new PanelResize());
    }
    
    public void enableInput(boolean val) {
               //remove and add so that there is only one 
              //instance of mouseClicks attached
    if(val) {
        removeMouseListener(mouseClicks);
        addMouseListener(mouseClicks);
    }
    else 
        removeMouseListener(mouseClicks);
    }

    public void update(Vector points) {
    float meanX, meanY;
    
    this.points = points;
    meanX = meanY = 0f;
    Enumeration pointsE = points.elements();
    while(pointsE.hasMoreElements()) {
        PointPair pair = (PointPair) pointsE.nextElement();
        meanX += pair.second.x;
        meanY += pair.second.y;
    }
            //point array may be empty
    meanX = points.size()==0 ? 0 : meanX/points.size();
    meanY = points.size()==0 ? 0 : meanY/points.size();

    centerOfGravity = new Point2D(meanX, meanY);
    xAxis = new Point2D(centerOfGravity.x + AXIS_SIZE,
                         centerOfGravity.y);
    yAxis = new Point2D(centerOfGravity.x, centerOfGravity.y - AXIS_SIZE);
    repaint();
    }
    
    /**
     * Overriding the update method performs double buffering, removing
     * flicker when points are moved.
     */

    public void update(Graphics g) {
    Graphics offScreenG = offScreenImage.getGraphics();
         
         //clear the offscreen area to the background color
    offScreenG.setColor(getBackground());
    offScreenG.fillRect( 0, 0, getSize().width, getSize().height );
         //paint on the offscreen image
    paint(offScreenG);
         //update the screen
    g.drawImage(offScreenImage,0,0,this);
    }

    
    /**
     * Draw the two point sets, the reference and transformed.
     * The reference points are circles and the transformed data are Xs.
     * Fiducials are red and target points are green.
     * A coordinate axis, manipulator is drawn in black. 
     * The manipulator will allow translation and rotation of the transformed 
     * point set.
     */  
    public void paint(Graphics g) {
    float x1, x2, y1, y2;
         //if no elements don't paint anything, special case so
         //we don't paint the coordinate axis
    if(points==null || points.isEmpty())
        return;

    Color c;
    Enumeration pointsE = points.elements();
    while(pointsE.hasMoreElements()) {
        PointPair pair = (PointPair) pointsE.nextElement();
    
              //draw a square for each of the original points,
             //fiducials are red, targets are green 
        c = pair.target == true ? Color.green : Color.red;
        g.setColor(c);
        final int PNT_NUM = 4;
        int xCoords[] = new int[PNT_NUM];
        int yCoords[] = new int[PNT_NUM];
        xCoords[0] = Math.round(pair.first.x - HALF_MARK_SIZE);
        xCoords[1] = Math.round(pair.first.x - HALF_MARK_SIZE);
        xCoords[2] = Math.round(pair.first.x + HALF_MARK_SIZE);
        xCoords[3] = Math.round(pair.first.x + HALF_MARK_SIZE);        
        yCoords[0] = Math.round(pair.first.y - HALF_MARK_SIZE);                
        yCoords[1] = Math.round(pair.first.y + HALF_MARK_SIZE);
        yCoords[2] = Math.round(pair.first.y + HALF_MARK_SIZE);        
        yCoords[3] = Math.round(pair.first.y - HALF_MARK_SIZE);
        g.drawPolygon(xCoords,yCoords,PNT_NUM);
              //draw an 'X' for each of the transformed points,
             //fiducials are red, targets are green 
        c = pair.target == true ? Color.green : Color.red;
        g.setColor(c);
        x1 = pair.second.x - HALF_MARK_SIZE;
        x2 = x1 + MARK_SIZE;
        y1 = pair.second.y - HALF_MARK_SIZE;
        y2 = y1 + MARK_SIZE;
        g.drawLine(Math.round(x1), Math.round(y1),
               Math.round(x2), Math.round(y2));
        g.drawLine(Math.round(x2), Math.round(y1),
               Math.round(x1), Math.round(y2));
    }
         //draw coordinate axis
    g.setColor(Color.black);

    g.fillOval(Math.round(centerOfGravity.x - HALF_AXIS_CIRCLE_SIZE),
           Math.round(centerOfGravity.y - HALF_AXIS_CIRCLE_SIZE),
           AXIS_CIRCLE_SIZE,
           AXIS_CIRCLE_SIZE);

    g.fillOval(Math.round(xAxis.x - HALF_AXIS_CIRCLE_SIZE),
           Math.round(xAxis.y - HALF_AXIS_CIRCLE_SIZE),
           AXIS_CIRCLE_SIZE,
           AXIS_CIRCLE_SIZE);

    g.fillOval(Math.round(yAxis.x - HALF_AXIS_CIRCLE_SIZE),
           Math.round(yAxis.y - HALF_AXIS_CIRCLE_SIZE),
           AXIS_CIRCLE_SIZE,
           AXIS_CIRCLE_SIZE);

    g.drawLine(Math.round(centerOfGravity.x), 
           Math.round(centerOfGravity.y),
           Math.round(xAxis.x), 
           Math.round(xAxis.y));
    g.drawLine(Math.round(centerOfGravity.x), 
           Math.round(centerOfGravity.y),
           Math.round(yAxis.x), 
           Math.round(yAxis.y));
    }
    
/*****************************************************************************
 * A class which sets the image size used for double buffering.
 * Whenever the panel is resized the double buffer image should be resized
 * too.
 */    
    class PanelResize extends ComponentAdapter {
    public void componentResized(ComponentEvent e) {
        Component c = e.getComponent();
        offScreenImage = createImage(c.getSize().width,c.getSize().height);
    }
    }
/*****************************************************************************
 * Two classes in charge of the mouse interaction. Addition of points and
 * manipulation of them (rotation and translation).
 *****************************************************************************/
    //This class should have extended MouseInputAdapter from Swing but I 
    //didn't want to use Swing here so I have to extend and implement.
    class MouseInputDrags extends MouseAdapter implements MouseMotionListener {
    private boolean cogDragged;
    private boolean xAxisDragged;
    private boolean yAxisDragged;

    public void mouseReleased(MouseEvent e) {
        if(cogDragged)
        cogDragged = false;
        else if(xAxisDragged)
        xAxisDragged = false;
        else if(yAxisDragged)
        yAxisDragged = false;
    }
    
        //Implement the two MouseMotionListener methods,
    //mouseMoved and mouseDragged 
    public void mouseMoved(MouseEvent e){}
    public void mouseDragged(MouseEvent e) {
        float dx, dy;
        int x, y;
        float x1, y1, x2, y2;

        x = e.getPoint().x;
        y = e.getPoint().y;
         //check if going to translate (user drags COG)
        dx = x - centerOfGravity.x;
        dy = y - centerOfGravity.y;
        if((cogDragged || 
           ((dx*dx+dy*dy) < AXIS_CIRCLE_SIZE*AXIS_CIRCLE_SIZE/4)) &&
           !xAxisDragged && !yAxisDragged) {
             //disallow action on clicks once the point 
             //set is manipulated
        Point2DManipulationPanel.this.enableInput(false);
        cogDragged = true;
        performTranslation(dx,dy);
        Point2DManipulationPanel.this.repaint();
        return;
        }
         //check if going to rotate (user drags yAxis)
        dx = x - yAxis.x;
        dy = y - yAxis.y;
        if((yAxisDragged || 
           ((dx*dx+dy*dy) < AXIS_CIRCLE_SIZE*AXIS_CIRCLE_SIZE/4)) &&
           !cogDragged && !xAxisDragged) {
             //disallow action on clicks once the point 
             //set is manipulated
        Point2DManipulationPanel.this.enableInput(false);
        yAxisDragged = true;
        x1 = x - centerOfGravity.x;
        y1 = centerOfGravity.y - y;
        x2 = yAxis.x - centerOfGravity.x;
        y2 = centerOfGravity.y - yAxis.y;
        performRotation(Math.atan2(y1*x2 - x1*y2,
                       x1*x2 + y1*y2));
        Point2DManipulationPanel.this.repaint();
        return;
        }
         //check if going to rotate (user drags xAxis)
        dx = x - xAxis.x;
        dy = y - xAxis.y;
        if((xAxisDragged || 
           ((dx*dx+dy*dy) < AXIS_CIRCLE_SIZE*AXIS_CIRCLE_SIZE/4)) &&
           !cogDragged && !yAxisDragged) {
             //disallow action on clicks once the point 
             //set is manipulated
        Point2DManipulationPanel.this.enableInput(false);
        xAxisDragged = true;
        x1 = x - centerOfGravity.x;
        y1 = centerOfGravity.y - y;
        x2 = xAxis.x - centerOfGravity.x;
        y2 = centerOfGravity.y - xAxis.y;
        performRotation(Math.atan2(y1*x2 - x1*y2,
                       x1*x2 + y1*y2));
        Point2DManipulationPanel.this.repaint();
        return;
        }
    }
    private void performTranslation(float dx, float dy) {
        centerOfGravity.x+=dx;
        centerOfGravity.y+=dy;
        yAxis.x+=dx;
        yAxis.y+=dy;
        xAxis.x+=dx;
        xAxis.y+=dy;
        Enumeration pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        PointPair pair = (PointPair) pointsE.nextElement();
        pair.second.x+=dx;
        pair.second.y+=dy;
        }
    }

    private void performRotation(double angle) {
        float locX, locY;
        float sTheta = (float)Math.sin(angle);
        float cTheta = (float)Math.cos(angle);
        locX = yAxis.x - centerOfGravity.x;
        locY = centerOfGravity.y - yAxis.y;
        yAxis.x = cTheta*locX- 
              sTheta*locY+
              centerOfGravity.x;
        yAxis.y = centerOfGravity.y - 
              sTheta*locX - 
              cTheta*locY;

        locX = xAxis.x - centerOfGravity.x;
        locY = centerOfGravity.y - xAxis.y;
        xAxis.x = cTheta*locX- 
              sTheta*locY+
              centerOfGravity.x;
        xAxis.y = centerOfGravity.y - 
              sTheta*locX - 
              cTheta*locY;
        Enumeration pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        PointPair pair = (PointPair) pointsE.nextElement();
        locX = pair.second.x - centerOfGravity.x;
        locY = centerOfGravity.y - pair.second.y;
        pair.second.x = cTheta*locX- 
                        sTheta*locY+
                        centerOfGravity.x;
        pair.second.y = centerOfGravity.y - 
                        sTheta*locX - 
                        cTheta*locY;
        }
    }
    }
    

    class MouseInputClicks extends MouseAdapter {

    public void mouseClicked(MouseEvent e) {
        int x, y, newCogX, newCogY;

        x = e.getPoint().x;
        y = e.getPoint().y;
        
        if(!e.isControlDown()) {//CTRL isn't pressed
        int n = points.size();
        centerOfGravity.setLocation((centerOfGravity.x*n + x)/(n+1),
                        (centerOfGravity.y*n + y)/(n+1));;
        yAxis.setLocation(centerOfGravity.x,        
                  centerOfGravity.y - AXIS_SIZE);
        xAxis.setLocation(centerOfGravity.x + AXIS_SIZE,        
                  centerOfGravity.y);
        points.addElement(new PointPair(new Point2D(x,y), 
                     new Point2D(x,y)));
        }  //check/uncheck if this is a selection of the point as a target
        else { 
        float dx, dy;
        Enumeration pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
            PointPair pair = (PointPair) pointsE.nextElement();
            dx = pair.first.x - x;
            dy = pair.first.y - y;
            if((dx*dx + dy*dy) < (MARK_SIZE*MARK_SIZE/4))
            pair.target = !pair.target;
        }
        }
        Point2DManipulationPanel.this.repaint();
    }
    }
}
