import Copyright.Copyright;
import java.util.*;
import java.awt.event.*;
import java.awt.*;

/**
 * An object displaying aspects of rigid registration, interacts with the data displayed in the 
 * Point2DManipulationPanel. Enables clearing, adding noise, adding bias, and registering the 2D point sets.
 * @author Ziv Yaniv (zivy@georgetown.edu)
 */
public class RigidRegistration2D extends Frame {
    private Vector points;     //pairs of points, reference and transfromed
    private Point2DManipulationPanel manipPanel;

    public RigidRegistration2D() {
    super("2D Rigid Registration");
    points = new Vector();
    manipPanel =  new Point2DManipulationPanel();
    manipPanel.update(points);
    add(manipPanel,BorderLayout.CENTER);
    Panel p = new Panel();
    add(createDataManipulationPanel(),BorderLayout.SOUTH);

    addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent e) {
            setVisible(false);
        }
        });
    
    }

    Panel createDataManipulationPanel() {
    Panel p = new Panel();

    Button b;
    b = new Button("Register");
    b.addActionListener(new RegisterActionListener());
    p.add(b);

    b = new Button("Clear");
    b.addActionListener(new ClearActionListener());
    p.add(b);

    b = new Button("Add noise");
    b.addActionListener(new AddNoiseActionListener());
    p.add(b);

    b = new Button("Add bias (FRE<TRE)");
    b.addActionListener(new AddBiasFRELessTREActionListener());
    p.add(b);

    b = new Button("Add bias (FRE>TRE)");
    b.addActionListener(new AddBiasFREGreaterTREActionListener());
    p.add(b);

    return p;
    }
    /**
     * Class attached to the 'Clear' menu button, removes all the
     * points.
     */
    private class ClearActionListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
        points.removeAllElements();
        manipPanel.update(points);
        manipPanel.enableInput(true);
    }
    }

    /**
     * Class attached to the 'Add noise' menu button. Adds random noise
     * to the 2D points in the range [-5,5] to the X and Y coordinates.
     */
    private class AddNoiseActionListener implements ActionListener {
    
    private static final int MAX_DX = 5;
    private static final int MAX_DY = 5;

    public void actionPerformed(ActionEvent e) {
        Random r = new Random();
        int sign;
        Enumeration pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        PointPair pair = (PointPair) pointsE.nextElement();
        if(!pair.target) {
            sign = r.nextInt() > 0 ? 1 : -1;  
            pair.second.x += MAX_DX*r.nextFloat()*sign;
            sign = r.nextInt() > 0 ? 1 : -1;  
            pair.second.y += MAX_DY*r.nextFloat()*sign;
        }
        }
        manipPanel.update(points);
        manipPanel.enableInput(false);
    }
    }

    /**
     * Class attached to the 'Add bias FRE<TRE' menu button. 
     * Adds X_BIAS and Y_BIAS to every fiducial point's coordinates.
     */
    private class AddBiasFRELessTREActionListener implements ActionListener {
    private static final int X_BIAS = 10;
    private static final int Y_BIAS = 10;

    public void actionPerformed(ActionEvent e) {
        Enumeration pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        PointPair pair = (PointPair) pointsE.nextElement();
        if(!pair.target) {
            pair.second.x += X_BIAS;
            pair.second.y += Y_BIAS;
        }
        }
        manipPanel.update(points);
        manipPanel.enableInput(false);
    }
    }
    /**
     * Class attached to the 'Add bias FRE>TRE' menu button. 
     * Adds +X_BIAS and +Y_BIAS to half of the fiducial point's and
     * -X_BIAS -Y_BIAS to the other half.
     */
    private class AddBiasFREGreaterTREActionListener 
    implements ActionListener {
    private static final int X_BIAS = 10;
    private static final int Y_BIAS = 10;

    public void actionPerformed(ActionEvent e) {
        Enumeration pointsE = points.elements();
                    //find out how many fiducials there are
        int fiducialNum=0;
        while(pointsE.hasMoreElements()) {
        if(!((PointPair)pointsE.nextElement()).target)
            fiducialNum++;
        }
        fiducialNum/=2;
                //add the bias accroding to the description above
        int pntNum = 0;
        pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        PointPair pair = (PointPair) pointsE.nextElement();
        if(!pair.target) {
            pntNum++;
            if(pntNum <= fiducialNum) {
            pair.second.x += X_BIAS;
            pair.second.y += Y_BIAS;
            }
            else {
            pair.second.x -= X_BIAS;
            pair.second.y -= Y_BIAS;
            }
        }
        }
        manipPanel.update(points);
        manipPanel.enableInput(false);
    }
    }
    /**
     * Class attached to the 'Register' menu button. Compute the rigid 
     * registration and apply its result to all the points.
     */
    private class RegisterActionListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
        PointPair pair;
        RigidRegistrationAlgorithm algo = new RigidRegistrationAlgorithm();

                        //create a fiducials only vector, the vector 
                       //RigidRegistration2D.points contains both 
                       //fiducial and target points
        Vector fiducials = new Vector();
        Enumeration pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        pair = (PointPair) pointsE.nextElement();
        if(!pair.target)
            fiducials.addElement(new PointPair(pair));
        }
        EuclideanTransformation2D transformation = algo.apply(fiducials);
        pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        pair = (PointPair) pointsE.nextElement();
        transformation.apply(pair.second);
        }
        manipPanel.update(points);
    }
    }
}
