import Copyright.Copyright;
import java.util.*;
import java.awt.event.*;
import java.awt.*;
import javax.swing.JOptionPane;
import java.io.*;

/**
 * An application displaying aspects of rigid registration. Has the same funcitonality as the applet and adds the
 * ability to save and load point sets (not possible with applets). Has a different interface than the applet.
 * @author Ziv Yaniv (zivy@georgetown.edu)
 */
public class RigidRegistration2DApplication extends Frame {
    private Vector points;     //pairs of points, reference and transfromed
    private Point2DManipulationPanel manipPanel;

    public RigidRegistration2DApplication() {
    super("2D Rigid Registration");
    points = new Vector();
    manipPanel =  new Point2DManipulationPanel();
    manipPanel.update(points);
    add("Center",manipPanel);
    setMenuBar(createMenus());
    }

    /**
     * Create the menus and attach action listeners to all of them.
     */    
    private MenuBar createMenus() {
    int i;
    MenuBar mb = new MenuBar();
    Menu fileM = new Menu("File");
    Menu editM = new Menu("Edit");
    mb.add(fileM);
    mb.add(editM);
                //add the "File" menu
    MenuItem [] fileMenu = {
        new MenuItem("Load points"),
        new MenuItem("Save Points"),
        new MenuItem("Exit")};
    
    ActionListener [] fileMenuActions = {
        new LoadPointsActionListener(),
        new SavePointsActionListener(),
        new ExitActionListener()};
    
    for(i=0; i<fileMenu.length; i++) {
        fileM.add(fileMenu[i]);
        fileMenu[i].addActionListener(fileMenuActions[i]);
    }
    fileM.insert(new MenuItem("-"),2);

                //add the "Edit" menu
    MenuItem [] editMenu = {
        new MenuItem("Clear"),
        new MenuItem("Register"),
        new MenuItem("Add noise"),
        new MenuItem("Add bias (FRE<TRE)"),
        new MenuItem("Add bias (FRE>TRE)")};
    
    ActionListener [] editMenuActions = {
        new ClearActionListener(),
        new RegisterActionListener(),
        new AddNoiseActionListener(),
        new AddBiasFRELessTREActionListener(),
        new AddBiasFREGreaterTREActionListener()
        };

    for(i=0; i<editMenu.length; i++) {
        editM.add(editMenu[i]);
        editMenu[i].addActionListener(editMenuActions[i]);
    }
    editM.insert(new MenuItem("-"),2);

    return(mb);
    }
    /**
     * Class attached to the 'Load points' menu button. Loads the 2d points
     * from a file into which they have been serialized.
     */
    private class LoadPointsActionListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
        String fileName, fullPathName;

        FileDialog fd = new FileDialog(RigidRegistration2DApplication.this, 
                       "Load points",FileDialog.LOAD);
        fd.show();
        if((fileName = fd.getFile()) != null) {
        try {
            fullPathName = fd.getDirectory() + fileName;
            ObjectInputStream in = 
              new ObjectInputStream(new FileInputStream(fullPathName));
            RigidRegistration2DApplication.this.points = (Vector) in.readObject();
            in.close(); //flush and close output stream
            manipPanel.update(RigidRegistration2DApplication.this.points);
            manipPanel.enableInput(true);
        }catch(Exception except) {
            JOptionPane.showMessageDialog(RigidRegistration2DApplication.this,
                                   "Error loading file.",
                                   "",JOptionPane.ERROR_MESSAGE);
        }
        }
    }
    }

    /**
     * Class attached to the 'Save points' menu button. Saves the 2d points
     * to a file using serialization.
     */
    private class SavePointsActionListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
        String fileName, fullPathName;

        FileDialog fd = new FileDialog(RigidRegistration2DApplication.this, 
                       "Save points",FileDialog.SAVE);
        fd.show();
        if((fileName = fd.getFile()) != null) {
        try {
            fullPathName = fd.getDirectory() + fileName;
            ObjectOutputStream out = 
            new ObjectOutputStream(new FileOutputStream(fullPathName));
            out.writeObject(RigidRegistration2DApplication.this.points);
            out.close(); //flush and close output stream
        }catch(Exception except) {
            JOptionPane.showMessageDialog(RigidRegistration2DApplication.this,
                                   "Error saving file.",
                                   "",JOptionPane.ERROR_MESSAGE);
        }
        }
    }
    }
    /**
     * Class attached to the 'Exit' menu button.
     */
    private class ExitActionListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
        System.exit(0);
    }
    }
    /**
     * Class attached to the 'Clear' menu button, removes all the
     * points.
     */
    private class ClearActionListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
        points.clear();
        manipPanel.update(points);
        manipPanel.enableInput(true);
    }
    }

    /**
     * Class attached to the 'Add noise' menu button. Adds random noise
     * to the 2D points in the range [-5,5] to the X and Y coordinates.
     */
    private class AddNoiseActionListener implements ActionListener {
    
    private static final int MAX_DX = 5;
    private static final int MAX_DY = 5;

    public void actionPerformed(ActionEvent e) {
        Random r = new Random();
        int sign;
        Enumeration pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        PointPair pair = (PointPair) pointsE.nextElement();
        if(!pair.target) {
            sign = r.nextBoolean() ? 1 : -1;  
            pair.second.x += MAX_DX*r.nextFloat()*sign;
            sign = r.nextBoolean() ? 1 : -1;  
            pair.second.y += MAX_DY*r.nextFloat()*sign;
        }
        }
        manipPanel.update(points);
        manipPanel.enableInput(false);
    }
    }

    /**
     * Class attached to the 'Add bias FRE<TRE' menu button. 
     * Adds X_BIAS and Y_BIAS to every fiducial point's coordinates.
     */
    private class AddBiasFRELessTREActionListener implements ActionListener {
    private static final int X_BIAS = 10;
    private static final int Y_BIAS = 10;

    public void actionPerformed(ActionEvent e) {
        Enumeration pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        PointPair pair = (PointPair) pointsE.nextElement();
        if(!pair.target) {
            pair.second.x += X_BIAS;
            pair.second.y += Y_BIAS;
        }
        }
        manipPanel.update(points);
        manipPanel.enableInput(false);
    }
    }
    /**
     * Class attached to the 'Add bias FRE>TRE' menu button. 
     * Adds +X_BIAS and +Y_BIAS to half of the fiducial point's and
     * -X_BIAS -Y_BIAS to the other half.
     */
    private class AddBiasFREGreaterTREActionListener 
    implements ActionListener {
    private static final int X_BIAS = 10;
    private static final int Y_BIAS = 10;

    public void actionPerformed(ActionEvent e) {
        Enumeration pointsE = points.elements();
                    //find out how many fiducials there are
        int fiducialNum=0;
        while(pointsE.hasMoreElements()) {
        if(!((PointPair)pointsE.nextElement()).target)
            fiducialNum++;
        }
        fiducialNum/=2;
                //add the bias accroding to the description above
        int pntNum = 0;
        pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        PointPair pair = (PointPair) pointsE.nextElement();
        if(!pair.target) {
            pntNum++;
            if(pntNum <= fiducialNum) {
            pair.second.x += X_BIAS;
            pair.second.y += Y_BIAS;
            }
            else {
            pair.second.x -= X_BIAS;
            pair.second.y -= Y_BIAS;
            }
        }
        }
        manipPanel.update(points);
        manipPanel.enableInput(false);
    }
    }
    /**
     * Class attached to the 'Register' menu button. Compute the rigid 
     * registration and apply its result to all the points.
     */
    private class RegisterActionListener implements ActionListener {
    public void actionPerformed(ActionEvent e) {
        PointPair pair;
        RigidRegistrationAlgorithm algo = new RigidRegistrationAlgorithm();

                        //create a fiducials only vector, the vector 
                       //RigidRegistration2DApplication.points contains both 
                       //fiducial and target points
        Vector fiducials = new Vector();
        Enumeration pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        pair = (PointPair) pointsE.nextElement();
        if(!pair.target)
            fiducials.add(new PointPair(pair));
        }
        EuclideanTransformation2D transformation = algo.apply(fiducials);
        pointsE = points.elements();
        while(pointsE.hasMoreElements()) {
        pair = (PointPair) pointsE.nextElement();
        transformation.apply(pair.second);
        }
        manipPanel.update(points);
    }
    }

    public static void main(String [] args) {
    final int WIDTH = 600;
    final int HEIGHT = 400;

    RigidRegistration2DApplication app = 
        new RigidRegistration2DApplication();
    app.addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent e) {
            System.exit(0);
        }
        });
    app.setSize(WIDTH,HEIGHT);
    app.setVisible(true);
    }
}
